from rerankers.models.ranker import BaseRanker
from rerankers.documents import Document


import torch
from typing import Union, List, Optional, Tuple
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
)
from rerankers.utils import (
    vprint,
    get_device,
    get_dtype,
    prep_docs,
)
from rerankers.results import RankedResults, Result


class TransformerRanker(BaseRanker):
    def __init__(
        self,
        model_name_or_path: str,
        dtype: Optional[Union[str, torch.dtype]] = None,
        device: Optional[Union[str, torch.device]] = None,
        batch_size: int = 16,
        verbose: int = 1,
        **kwargs,
    ):
        self.verbose = verbose
        self.device = get_device(device, verbose=self.verbose)
        self.dtype = get_dtype(dtype, self.device, self.verbose)
        model_kwargs = kwargs.get("model_kwargs", {})
        self.model = AutoModelForSequenceClassification.from_pretrained(
            model_name_or_path,
            torch_dtype=self.dtype,
            **model_kwargs,
        ).to(self.device)
        vprint(f"Loaded model {model_name_or_path}", self.verbose)
        vprint(f"Using device {self.device}.", self.verbose)
        vprint(f"Using dtype {self.dtype}.", self.verbose)
        self.model.eval()
        tokenizer_kwargs = kwargs.get("tokenizer_kwargs", {})
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path,
            **tokenizer_kwargs,
        )
        self.ranking_type = "pointwise"
        self.batch_size = batch_size

    def tokenize(self, inputs: Union[str, List[str], List[Tuple[str, str]]]):
        return self.tokenizer(
            inputs, return_tensors="pt", padding=True, truncation=True
        ).to(self.device)

    @torch.inference_mode()
    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
        batch_size: Optional[int] = None,
    ) -> RankedResults:
        docs = prep_docs(docs, doc_ids, metadata)
        inputs = [(query, doc.text) for doc in docs]

        # Override self.batch_size if explicitely set
        if batch_size is None:
            batch_size = self.batch_size
        batched_inputs = [
            inputs[i : i + batch_size] for i in range(0, len(inputs), batch_size)
        ]
        scores = []
        for batch in batched_inputs:
            tokenized_inputs = self.tokenize(batch)
            batch_scores = self.model(**tokenized_inputs).logits.squeeze()
            batch_scores = batch_scores.detach().cpu().numpy().tolist()
            if isinstance(batch_scores, float):  # Handling the case of single score
                scores.append(batch_scores)
            else:
                scores.extend(batch_scores)
        if len(scores) == 1:
            return Result(document=docs[0], score=scores[0])
        else:
            ranked_results = [
                Result(document=doc, score=score, rank=idx + 1)
                for idx, (doc, score) in enumerate(
                    sorted(zip(docs, scores), key=lambda x: x[1], reverse=True)
                )
            ]
            return RankedResults(results=ranked_results, query=query, has_scores=True)

    @torch.inference_mode()
    def score(self, query: str, doc: str) -> float:
        inputs = self.tokenize((query, doc))
        outputs = self.model(**inputs)
        score = outputs.logits.squeeze().detach().cpu().numpy().astype(float)
        return score



from typing import Optional, Union, Literal
from pydantic import BaseModel, validator


class Document(BaseModel):
    document_type: Literal["text", "image"] = "text"
    text: Optional[str] = None
    base64: Optional[str] = None
    image_path: Optional[str] = None
    doc_id: Optional[Union[str, int]] = None
    metadata: Optional[dict] = None

    @validator("text")
    def validate_text(cls, v, values):
        if values.get("document_type") == "text" and v is None:
            raise ValueError("text field is required when document_type is 'text'")
        return v

    def __init__(
        self,
        text: Optional[str] = None,
        doc_id: Optional[Union[str, int]] = None,
        metadata: Optional[dict] = None,
        document_type: Literal["text", "image"] = "text",
        image_path: Optional[str] = None,
        base64: Optional[str] = None,
    ):
        super().__init__(
            text=text,
            doc_id=doc_id,
            metadata=metadata,
            document_type=document_type,
            base64=base64,
            image_path=image_path,
        )



import torch
from PIL import Image
import base64
import io
# TODO: Support more than Qwen
from transformers import AutoProcessor, Qwen2VLForConditionalGeneration
from rerankers.models.ranker import BaseRanker
from rerankers.documents import Document
from typing import Union, List, Optional
from rerankers.utils import vprint, get_device, get_dtype, prep_image_docs
from rerankers.results import RankedResults, Result

PREDICTION_TOKENS = {
    "default": ["False", "True"],
    "lightonai/MonoQwen2-VL-v0.1": ["False", "True"]
}

def _get_output_tokens(model_name_or_path, token_false: str, token_true: str):
    if token_false == "auto":
        if model_name_or_path in PREDICTION_TOKENS:
            token_false = PREDICTION_TOKENS[model_name_or_path][0]
        else:
            token_false = PREDICTION_TOKENS["default"][0]
            print(
                f"WARNING: Model {model_name_or_path} does not have known True/False tokens. Defaulting token_false to `{token_false}`."
            )
    if token_true == "auto":
        if model_name_or_path in PREDICTION_TOKENS:
            token_true = PREDICTION_TOKENS[model_name_or_path][1]
        else:
            token_true = PREDICTION_TOKENS["default"][1]
            print(
                f"WARNING: Model {model_name_or_path} does not have known True/False tokens. Defaulting token_true to `{token_true}`."
            )

    return token_false, token_true

class MonoVLMRanker(BaseRanker):
    def __init__(
        self,
        model_name_or_path: str,
        processor_name: Optional[str] = None,
        dtype: Optional[Union[str, torch.dtype]] = 'bf16',
        device: Optional[Union[str, torch.device]] = None,
        batch_size: int = 1,
        verbose: int = 1,
        token_false: str = "auto",
        token_true: str = "auto",
        return_logits: bool = False,
        prompt_template: str = "Assert the relevance of the previous image document to the following query, answer True or False. The query is: {query}",
        **kwargs
    ):
        self.verbose = verbose
        self.device = get_device(device, verbose=self.verbose)
        if self.device == 'mps':
            print("WARNING: MPS is not supported by MonoVLMRanker due to PyTorch limitations. Falling back to CPU.")
            self.device = 'cpu'
        print(dtype)
        self.dtype = get_dtype(dtype, self.device, self.verbose)
        self.batch_size = batch_size
        self.return_logits = return_logits
        self.prompt_template = prompt_template

        vprint(f"Loading model {model_name_or_path}, this might take a while...", self.verbose)
        vprint(f"Using device {self.device}.", self.verbose)
        vprint(f"Using dtype {self.dtype}.", self.verbose)

        processor_name = processor_name or "Qwen/Qwen2-VL-2B-Instruct"
        processor_kwargs = kwargs.get("processor_kwargs", {})
        model_kwargs = kwargs.get("model_kwargs", {})
        attention_implementation = kwargs.get("attention_implementation", "flash_attention_2")
        self.processor = AutoProcessor.from_pretrained(processor_name, **processor_kwargs)
        self.model = Qwen2VLForConditionalGeneration.from_pretrained(
            model_name_or_path,
            device_map=self.device,
            torch_dtype=self.dtype,
            attn_implementation=attention_implementation,
            **model_kwargs
        )
        self.model.eval()

        token_false, token_true = _get_output_tokens(
            model_name_or_path=model_name_or_path,
            token_false=token_false,
            token_true=token_true,
        )
        self.token_false_id = self.processor.tokenizer.convert_tokens_to_ids(token_false)
        self.token_true_id = self.processor.tokenizer.convert_tokens_to_ids(token_true)
        
        vprint(f"VLM true token set to {token_true}", self.verbose)
        vprint(f"VLM false token set to {token_false}", self.verbose)

    @torch.inference_mode()
    def _get_scores(self, query: str, docs: List[Document]) -> List[float]:
        scores = []
        for doc in docs:
            if doc.document_type != "image" or not doc.base64:
                raise ValueError("MonoVLMRanker requires image documents with base64 data")
            
            # Convert base64 to PIL Image
            image_io = io.BytesIO(base64.b64decode(doc.base64))
            image_io.seek(0)  # Reset file pointer to start
            image = Image.open(image_io).convert('RGB')

            # Prepare prompt
            prompt = self.prompt_template.format(query=query)
            messages = [
                {
                    "role": "user",
                    "content": [
                        {"type": "image", "image": image},
                        {"type": "text", "text": prompt},
                    ],
                }
            ]

            # Process inputs
            text = self.processor.apply_chat_template(
                messages, 
                tokenize=False, 
                add_generation_prompt=True
            )
            inputs = self.processor(
                text=text, 
                images=image, 
                return_tensors="pt"
            ).to(self.device).to(self.dtype)

            # Get model outputs
            outputs = self.model(**inputs)
            logits = outputs.logits[:, -1, :]
            
            # Calculate scores
            relevant_logits = logits[:, [self.token_false_id, self.token_true_id]]
            if self.return_logits:
                score = relevant_logits[0, 1].cpu().item()  # True logit
            else:
                probs = torch.softmax(relevant_logits, dim=-1)
                score = probs[0, 1].cpu().item()  # True probability
            
            scores.append(score)
            
        return scores

    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
    ) -> RankedResults:
        docs = prep_image_docs(docs, doc_ids, metadata)
        scores = self._get_scores(query, docs)
        ranked_results = [
            Result(document=doc, score=score, rank=idx + 1)
            for idx, (doc, score) in enumerate(
                sorted(zip(docs, scores), key=lambda x: x[1], reverse=True)
            )
        ]
        return RankedResults(results=ranked_results, query=query, has_scores=True)

    def score(self, query: str, doc: Union[str, Document]) -> float:
        scores = self._get_scores(query, [doc])
        return scores[0]



import torch
from transformers import AutoModelForCausalLM, AutoTokenizer
from rerankers.models.ranker import BaseRanker
from rerankers.documents import Document
from typing import Union, List, Optional
from rerankers.utils import vprint, get_device, get_dtype, prep_docs
from rerankers.results import RankedResults, Result


PROMPTS = {
    "BAAI/bge-reranker-v2.5-gemma2-lightweight": "Given a query A and a passage B, determine whether the passage contains an answer to the query by providing a prediction of either 'Yes' or 'No'.",
    "default": "Given a query A and a passage B, determine whether the passage contains an answer to the query by providing a prediction of either 'Yes' or 'No'.",
}

DEFAULT_PARAMS = {
    "default": {},
    "BAAI/bge-multilingual-gemma2": {},
    "BAAI/bge-reranker-v2-gemma": {},
    "BAAI/bge-reranker-v2-minicpm-layerwise": {"cutoff_layers": [28]},
    "BAAI/bge-reranker-v2.5-gemma2-lightweight": {
        "cutoff_layers": [28],
        "compress_ratio": 2,
        "compress_layer": [24, 40],
    },
}


class LLMLayerWiseRanker(BaseRanker):
    def __init__(
        self,
        model_name_or_path: str = "BAAI/bge-reranker-v2.5-gemma2-lightweight",
        max_sequence_length: int = 512,
        dtype: Optional[Union[str, torch.dtype]] = None,
        device: Optional[Union[str, torch.device]] = None,
        batch_size: int = 16,
        verbose: int = 1,
        prompt: Optional[str] = None,
        cutoff_layers: Optional[List[int]] = None,
        compress_ratio: Optional[int] = None,
        compress_layer: Optional[List[int]] = None,
        **kwargs,
    ):
        self.verbose = verbose
        self.device = get_device(device, verbose=self.verbose)
        self.dtype = get_dtype(dtype, self.device, self.verbose)
        self.batch_size = batch_size

        vprint(
            f"Loading model {model_name_or_path}, this might take a while...",
            self.verbose,
        )
        vprint(f"Using device {self.device}.", self.verbose)
        vprint(f"Using dtype {self.dtype}.", self.verbose)
        tokenizer_kwargs = kwargs.get("tokenizer_kwargs", {})
        tokenizer_trust_remote_code = tokenizer_kwargs.pop("trust_remote_code", True)
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_or_path,
            trust_remote_code=tokenizer_trust_remote_code,
            **tokenizer_kwargs,
        )
        self.max_sequence_length = max_sequence_length
        self.tokenizer.model_max_length = self.max_sequence_length
        self.tokenizer.padding_side = "right"
        model_kwargs = kwargs.get("model_kwargs", {})
        model_trust_remote_code = model_kwargs.pop("trust_remote_code", True)

        self.model = AutoModelForCausalLM.from_pretrained(
            model_name_or_path,
            trust_remote_code=model_trust_remote_code,
            torch_dtype=self.dtype,
            **model_kwargs,
        ).to(self.device)
        self.model.eval()

        # Create params dict based on specified values or defaults
        params = {}
        if cutoff_layers is not None:
            params["cutoff_layers"] = cutoff_layers
        if compress_ratio is not None:
            params["compress_ratio"] = compress_ratio
        if compress_layer is not None:
            params["compress_layer"] = compress_layer
        if not params:
            params = DEFAULT_PARAMS.get(model_name_or_path, DEFAULT_PARAMS["default"])
        self.params = params

        self.prompt = prompt
        if self.prompt is None:
            self.prompt = PROMPTS.get(model_name_or_path, PROMPTS["default"])

    def _get_inputs(self, pairs, max_sequence_length: int):
        prompt = self.prompt
        sep = "\n"
        prompt_inputs = self.tokenizer(
            prompt, return_tensors=None, add_special_tokens=False
        )["input_ids"]
        sep_inputs = self.tokenizer(sep, return_tensors=None, add_special_tokens=False)[
            "input_ids"
        ]
        inputs = []
        for query, passage in pairs:
            query_inputs = self.tokenizer(
                f"A: {query}",
                return_tensors=None,
                add_special_tokens=False,
                max_length=max_sequence_length * 3 // 4,
                truncation=True,
            )
            passage_inputs = self.tokenizer(
                f"B: {passage}",
                return_tensors=None,
                add_special_tokens=False,
                max_length=max_sequence_length,
                truncation=True,
            )
            item = self.tokenizer.prepare_for_model(
                [self.tokenizer.bos_token_id] + query_inputs["input_ids"],
                sep_inputs + passage_inputs["input_ids"],
                truncation="only_second",
                max_length=max_sequence_length,
                padding=False,
                return_attention_mask=False,
                return_token_type_ids=False,
                add_special_tokens=False,
            )
            item["input_ids"] = item["input_ids"] + sep_inputs + prompt_inputs
            item["attention_mask"] = [1] * len(item["input_ids"])
            inputs.append(item)

        return self.tokenizer.pad(
            inputs,
            padding=True,
            max_length=max_sequence_length + len(sep_inputs) + len(prompt_inputs),
            pad_to_multiple_of=8,
            return_tensors="pt",
        )

    @torch.inference_mode()
    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
        batch_size: Optional[int] = None,
        max_sequence_length: Optional[int] = None,
    ) -> RankedResults:
        docs = prep_docs(docs, doc_ids, metadata)
        pairs = [(query, doc.text) for doc in docs]

        # Override self.batch_size if explicitly set
        if batch_size is None:
            batch_size = self.batch_size

        # Same for max_sequence_length
        if max_sequence_length is None:
            max_sequence_length = self.max_sequence_length

        batched_pairs = [
            pairs[i : i + batch_size] for i in range(0, len(pairs), batch_size)
        ]
        scores = []

        for batch in batched_pairs:
            inputs = self._get_inputs(batch, max_sequence_length=max_sequence_length)
            inputs = {k: v.to(self.device) for k, v in inputs.items()}

            outputs = self.model(**inputs, return_dict=True, **self.params)
            all_scores = [
                scores[:, -1]
                .view(
                    -1,
                )
                .float()
                for scores in outputs[0]
            ]
            batch_scores = all_scores[-1].cpu().numpy().tolist()

            scores.extend(batch_scores)

        ranked_results = [
            Result(document=doc, score=score, rank=idx + 1)
            for idx, (doc, score) in enumerate(
                sorted(zip(docs, scores), key=lambda x: x[1], reverse=True)
            )
        ]
        return RankedResults(results=ranked_results, query=query, has_scores=True)

    @torch.inference_mode()
    def score(self, query: str, doc: str) -> float:
        inputs = self._get_inputs(
            [(query, doc)], max_sequence_length=self.max_sequence_length
        )
        inputs = {k: v.to(self.device) for k, v in inputs.items()}

        outputs = self.model(**inputs, return_dict=True, **self.params)
        all_scores = [
            scores[:, -1]
            .view(
                -1,
            )
            .float()
            for scores in outputs[0]
        ]
        score = all_scores[-1].item()

        return score



from unittest.mock import patch
import torch
from rerankers import Reranker
from rerankers.models.transformer_ranker import TransformerRanker
from rerankers.results import Result, RankedResults
from rerankers.documents import Document

@patch("rerankers.models.transformer_ranker.TransformerRanker.rank")
def test_transformer_ranker_rank(mock_rank):
    query = "Gone with the wind is an absolute masterpiece"
    docs = [
        "Gone with the wind is a masterclass in bad storytelling.",
        "Gone with the wind is an all-time classic",
    ]
    expected_results = RankedResults(
        results=[
            Result(
                document=Document(
                    doc_id=1, text="Gone with the wind is an all-time classic"
                ),
                score=1.6181640625,
                rank=1,
            ),
            Result(
                document=Document(
                    doc_id=0,
                    text="Gone with the wind is a masterclass in bad storytelling.",
                ),
                score=0.88427734375,
                rank=2,
            ),
        ],
        query=query,
        has_scores=True,
    )
    mock_rank.return_value = expected_results
    ranker = TransformerRanker("mixedbread-ai/mxbai-rerank-xsmall-v1")
    results = ranker.rank(query=query, docs=docs)
    assert results == expected_results



import base64
import binascii
from typing import Union, Optional, List, Iterable
try:
    import io
    from PIL import Image
except ImportError:
    pass
from rerankers.documents import Document

def vprint(txt: str, verbose: int) -> None:
    if verbose > 0:
        print(txt)


try:
    import torch

    def get_dtype(
        dtype: Optional[Union[str, torch.dtype]],
        device: Optional[Union[str, torch.device]],
        verbose: int = 1,
    ) -> torch.dtype:
        if dtype is None:
            vprint("No dtype set", verbose)
        # if device == "cpu":
        #     vprint("Device set to `cpu`, setting dtype to `float32`", verbose)
        #     dtype = torch.float32
        if not isinstance(dtype, torch.dtype):
            if dtype == "fp16" or dtype == "float16":
                dtype = torch.float16
            elif dtype == "bf16" or dtype == "bfloat16":
                dtype = torch.bfloat16
            else:
                dtype = torch.float32
        vprint(f"Using dtype {dtype}", verbose)
        return dtype

    def get_device(
        device: Optional[Union[str, torch.device]],
        verbose: int = 1,
        no_mps: bool = False,
    ) -> Union[str, torch.device]:
        if not device:
            vprint("No device set", verbose)
            if torch.cuda.is_available():
                device = "cuda"
            elif torch.backends.mps.is_available() and not no_mps:
                device = "mps"
            else:
                device = "cpu"
            vprint(f"Using device {device}", verbose)
        return device

except ImportError:
    pass


def make_documents(
    docs: List[str],
    doc_ids: Optional[Union[List[str], List[int]]] = None,
):
    if doc_ids is None:
        doc_ids = list(range(len(docs)))
    return [Document(doc, doc_id=doc_ids[i]) for i, doc in enumerate(docs)]


def prep_docs(
    docs: Union[str, List[str], Document, List[Document]],
    doc_ids: Optional[Union[List[str], List[int]]] = None,
    metadata: Optional[List[dict]] = None,
):
    if isinstance(docs, Document) or (
        isinstance(docs, List) and isinstance(docs[0], Document)
    ):
        if isinstance(docs, Document):
            docs = [docs]
        if doc_ids is not None:
            if docs[0].doc_id is not None:
                print(
                    "Overriding doc_ids passed within the Document objects with explicitly passed doc_ids!"
                )
                print(
                    "This is not the preferred way of doing so, please double-check your code."
                )
            for i, doc in enumerate(docs):
                doc.doc_id = doc_ids[i]

        elif doc_ids is None:
            doc_ids = [doc.doc_id for doc in docs]
            if doc_ids[0] is None:
                print(
                    "'None' doc_ids detected, reverting to auto-generated integer ids..."
                )
                doc_ids = list(range(len(docs)))

        if metadata is not None:
            if docs[0].meatadata is not None:
                print(
                    "Overriding doc_ids passed within the Document objects with explicitly passed doc_ids!"
                )
                print(
                    "This is not the preferred way of doing so, please double-check your code."
                )
            for i, doc in enumerate(docs):
                doc.metadata = metadata[i]

        return docs

    if isinstance(docs, str):
        docs = [docs]
    if doc_ids is None:
        doc_ids = list(range(len(docs)))
    if metadata is None:
        metadata = [{} for _ in docs]

    return [
        Document(doc, doc_id=doc_ids[i], metadata=metadata[i])
        for i, doc in enumerate(docs)
    ]


def prep_image_docs(
    docs: Union[str, List[str], Document, List[Document]],
    doc_ids: Optional[Union[List[str], List[int]]] = None,
    metadata: Optional[List[dict]] = None,
) -> List[Document]:
    """
    Prepare image documents for processing. Can handle base64 encoded images or file paths.
    Similar to prep_docs but specialized for image documents.
    """
    # If already Document objects, handle similarly to prep_docs
    if isinstance(docs, Document) or (
        isinstance(docs, List) and isinstance(docs[0], Document)
    ):
        if isinstance(docs, Document):
            docs = [docs]
        # Validate all docs are image type
        for doc in docs:
            if doc.document_type != "image":
                raise ValueError("All documents must be of type 'image'")
        return prep_docs(docs, doc_ids, metadata)

    # Handle string inputs (paths or base64)
    if isinstance(docs, str):
        docs = [docs]

    processed_docs = []
    for doc in docs:
        # Check if input is base64 by attempting to decode
        try:
            # Try to decode and verify it's an image
            decoded = base64.b64decode(doc)
            try:
                Image.open(io.BytesIO(decoded)).verify()
                b64 = doc
                image_path = None
            except:
                raise binascii.Error("Invalid image data")
        except binascii.Error:
            # If decode fails, treat as file path
            try:
                image_path = doc
                with open(doc, 'rb') as img_file:
                    b64 = base64.b64encode(img_file.read()).decode('utf-8')
            except Exception as e:
                raise ValueError(f"Could not process image input {doc}: {str(e)}")
        
        processed_docs.append(
            Document(
                document_type="image",
                base64=b64,
                image_path=image_path
            )
        )

    # Handle doc_ids and metadata
    if doc_ids is None:
        doc_ids = list(range(len(processed_docs)))
    if metadata is None:
        metadata = [{} for _ in processed_docs]

    # Set doc_ids and metadata
    for i, doc in enumerate(processed_docs):
        doc.doc_id = doc_ids[i]
        doc.metadata = metadata[i]


    return processed_docs




def get_chunks(iterable: Iterable, chunk_size: int):  # noqa: E741
    """
    Implementation from https://github.com/unicamp-dl/InRanker/blob/main/inranker/base.py with extra typing and more descriptive names.
    This method is used to split a list l into chunks of batch size n.
    """
    for i in range(0, len(iterable), chunk_size):
        yield iterable[i : i + chunk_size]



from rerankers.models.ranker import BaseRanker

from flashrank import Ranker, RerankRequest


from typing import Union, List, Optional, Tuple
from rerankers.utils import vprint, prep_docs
from rerankers.results import RankedResults, Result
from rerankers.documents import Document


class FlashRankRanker(BaseRanker):
    def __init__(
        self,
        model_name_or_path: str,
        verbose: int = 1,
        cache_dir: str = "./.flashrank_cache",
    ):
        self.verbose = verbose
        vprint(
            f"Loading model FlashRank model {model_name_or_path}...", verbose=verbose
        )
        self.model = Ranker(model_name=model_name_or_path, cache_dir=cache_dir)
        self.ranking_type = "pointwise"

    def tokenize(self, inputs: Union[str, List[str], List[Tuple[str, str]]]):
        return self.tokenizer(
            inputs, return_tensors="pt", padding=True, truncation=True
        ).to(self.device)

    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
    ) -> RankedResults:
        docs = prep_docs(docs, doc_ids, metadata)
        passages = [
            {"id": doc_idx, "text": doc.text} for doc_idx, doc in enumerate(docs)
        ]

        rerank_request = RerankRequest(query=query, passages=passages)
        flashrank_results = self.model.rerank(rerank_request)

        ranked_results = [
            Result(
                document=docs[result["id"]],  # Returns reranked documents.
                score=result["score"],
                rank=idx + 1,
            )
            for idx, result in enumerate(flashrank_results)
        ]
        
        return RankedResults(results=ranked_results, query=query, has_scores=True)

    def score(self, query: str, doc: str) -> float:
        rerank_request = RerankRequest(
            query=query, passages=[{"id": "temp_id", "text": doc}]
        )
        flashrank_result = self.model.rerank(rerank_request)
        score = flashrank_result[0]["score"]
        return score



from typing import Union, List, Optional
from rerankers.models.ranker import BaseRanker
from rerankers.results import RankedResults, Result
from rerankers.utils import prep_docs
from rerankers.documents import Document
from string import Template


import requests
import json


URLS = {
    "cohere": "https://api.cohere.ai/v1/rerank",
    "jina": "https://api.jina.ai/v1/rerank",
    "voyage": "https://api.voyageai.com/v1/rerank",
    "mixedbread.ai": "https://api.mixedbread.ai/v1/reranking",
    "pinecone": "https://api.pinecone.io/rerank",
}
AUTHORIZATION_KEY_MAPPING = {
    "pinecone": "Api-Key"
}
API_VERSION_MAPPING = {
    "pinecone": {"X-Pinecone-API-Version": "2024-10"}
}

API_KEY_MAPPING = {
    "pinecone": Template("$api_key")
}

DOCUMENT_KEY_MAPPING = {
    "mixedbread.ai": "input",
    "text-embeddings-inference":"texts"
}
RETURN_DOCUMENTS_KEY_MAPPING = {
    "mixedbread.ai":"return_input",
    "text-embeddings-inference":"return_text"
}
RESULTS_KEY_MAPPING = {
    "voyage": "data",
    "mixedbread.ai": "data",
    "pinecone": "data",
    "text-embeddings-inference": None
}
SCORE_KEY_MAPPING = {
    "mixedbread.ai": "score",
    "pinecone": "score",
    "text-embeddings-inference":"score"
}

class APIRanker(BaseRanker):
    def __init__(self, model: str, api_key: str, api_provider: str, verbose: int = 1, url: str = None):
        self.api_provider = api_provider.lower()
        self.api_key = API_KEY_MAPPING.get(self.api_provider,Template("Bearer $api_key")).substitute(api_key=api_key)
        authorization_key = AUTHORIZATION_KEY_MAPPING.get(self.api_provider,"Authorization")
        api_version = API_VERSION_MAPPING.get(self.api_provider,None)
        self.model = model
        self.verbose = verbose
        self.ranking_type = "pointwise"
        self.headers = {
            "accept": "application/json",
            "content-type": "application/json",
            authorization_key: self.api_key,
        }
        if api_version:
            self.headers.update(api_version)
        self.url = url if url else URLS[self.api_provider]


    def _get_document_text(self, r: dict) -> str:
        if self.api_provider == "voyage":
            return r["document"]
        elif self.api_provider == "mixedbread.ai":
            return r["input"]
        elif self.api_provider == "text-embeddings-inference":
            return r["text"]
        else:
            return r["document"]["text"]

    def _get_score(self, r: dict) -> float:
        score_key = SCORE_KEY_MAPPING.get(self.api_provider,"relevance_score")
        return r[score_key]

    def _parse_response(
        self, response: dict,  docs: List[Document],
    ) -> RankedResults:
        ranked_docs = []
        results_key = RESULTS_KEY_MAPPING.get(self.api_provider,"results")

        for i, r in enumerate(response[results_key] if results_key else response):
            ranked_docs.append(
                Result(
                    document=docs[r["index"]],
                    text=self._get_document_text(r),
                    score=self._get_score(r),
                    rank=i + 1,
                )
            )

        return ranked_docs

    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
    ) -> RankedResults:
        docs = prep_docs(docs, doc_ids, metadata)
        payload = self._format_payload(query, docs)
        response = requests.post(self.url, headers=self.headers, data=payload)
        results = self._parse_response(response.json(), docs)
        return RankedResults(results=results, query=query, has_scores=True)


    def _format_payload(self, query: str, docs: List[str]) -> str:
        top_key = (
            "top_n" if self.api_provider not in ["voyage", "mixedbread.ai"] else "top_k"
        )
        documents_key = DOCUMENT_KEY_MAPPING.get(self.api_provider,"documents")
        return_documents_key = RETURN_DOCUMENTS_KEY_MAPPING.get(self.api_provider,"return_documents")

        documents = (
            [d.text for d in docs] if self.api_provider not in ["pinecone"] else [{"text": d.text} for d in docs]
        ) 
        
        payload = {
            "model": self.model,
            "query": query,
            documents_key: documents,
            top_key: len(docs),
            return_documents_key: True,
        }
        return json.dumps(payload)

    def score(self, query: str, doc: str) -> float:
        payload = self._format_payload(query, [doc])
        response = requests.post(self.url, headers=self.headers, data=payload)
        results = self._parse_response(response.json(), [doc])
        return results[0].score



"""Code from HotchPotch's JQaRa repository: https://github.com/hotchpotch/JQaRA/blob/main/evaluator/reranker/colbert_reranker.py
Modifications include packaging into a BaseRanker, dynamic query/doc length and batch size handling."""

import torch
import torch.nn as nn
from transformers import BertPreTrainedModel, BertModel, AutoModel, AutoTokenizer
from typing import List, Optional, Union
from math import ceil

from rerankers.models.ranker import BaseRanker
from rerankers.documents import Document
from rerankers.results import RankedResults, Result
from rerankers.utils import vprint, get_device, get_dtype, prep_docs


def _insert_token(
    output: dict,
    insert_token_id: int,
    insert_position: int = 1,
    token_type_id: int = 0,
    attention_value: int = 1,
):
    """
    Inserts a new token at a specified position into the sequences of a tokenized representation.

    This function takes a dictionary containing tokenized representations
    (e.g., 'input_ids', 'token_type_ids', 'attention_mask') as PyTorch tensors,
    and inserts a specified token into each sequence at the given position.
    This can be used to add special tokens or other modifications to tokenized inputs.

    Parameters:
    - output (dict): A dictionary containing the tokenized representations. Expected keys
                     are 'input_ids', 'token_type_ids', and 'attention_mask'. Each key
                     is associated with a PyTorch tensor.
    - insert_token_id (int): The token ID to be inserted into each sequence.
    - insert_position (int, optional): The position in the sequence where the new token
                                       should be inserted. Defaults to 1, which typically
                                       follows a special starting token like '[CLS]' or '[BOS]'.
    - token_type_id (int, optional): The token type ID to assign to the inserted token.
                                     Defaults to 0.
    - attention_value (int, optional): The attention mask value to assign to the inserted token.
                                        Defaults to 1.

    Returns:
    - updated_output (dict): A dictionary containing the updated tokenized representations,
                             with the new token inserted at the specified position in each sequence.
                             The structure and keys of the output dictionary are the same as the input.
    """
    updated_output = {}
    for key in output:
        updated_tensor_list = []
        for seqs in output[key]:
            if len(seqs.shape) == 1:
                seqs = seqs.unsqueeze(0)
            for seq in seqs:
                first_part = seq[:insert_position]
                second_part = seq[insert_position:]
                new_element = (
                    torch.tensor([insert_token_id])
                    if key == "input_ids"
                    else torch.tensor([token_type_id])
                )
                if key == "attention_mask":
                    new_element = torch.tensor([attention_value])
                updated_seq = torch.cat((first_part, new_element, second_part), dim=0)
                updated_tensor_list.append(updated_seq)
        updated_output[key] = torch.stack(updated_tensor_list)
    return updated_output


def _colbert_score(q_reps, p_reps, q_mask: torch.Tensor, p_mask: torch.Tensor):
    # calc max sim
    # base code from: https://github.com/FlagOpen/FlagEmbedding/blob/master/FlagEmbedding/BGE_M3/modeling.py

    # Assert that all q_reps are at least as long as the query length
    assert (
        q_reps.shape[1] >= q_mask.shape[1]
    ), f"q_reps should have at least {q_mask.shape[1]} tokens, but has {q_reps.shape[1]}"

    token_scores = torch.einsum("qin,pjn->qipj", q_reps, p_reps)
    token_scores = token_scores.masked_fill(p_mask.unsqueeze(0).unsqueeze(0) == 0, -1e4)
    scores, _ = token_scores.max(-1)
    scores = scores.sum(1) / q_mask.sum(-1, keepdim=True)
    return scores


class ColBERTModel(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.bert = BertModel(config)

        # TODO: Load from artifact.metadata
        if "small" in config._name_or_path:
            linear_dim = 96
        else:
            linear_dim = 128
        print(f"Linear Dim set to: {linear_dim} for downcasting")
        self.linear = nn.Linear(config.hidden_size, linear_dim, bias=False)
        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        encoder_hidden_states=None,
        encoder_attention_mask=None,
        output_attentions=None,
        output_hidden_states=None,
    ):
        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=True,  # Always output hidden states
        )

        sequence_output = outputs[0]

        return self.linear(sequence_output)

    def _encode(self, texts: list[str], insert_token_id: int, is_query: bool = False):
        encoding = self.tokenizer(
            texts,
            return_tensors="pt",
            padding=True,
            max_length=self.max_length - 1,  # for insert token
            truncation=True,
        )
        encoding = _insert_token(encoding, insert_token_id)  # type: ignore

        if is_query:
            mask_token_id = self.tokenizer.mask_token_id

            new_encodings = {"input_ids": [], "attention_mask": []}

            for i, input_ids in enumerate(encoding["input_ids"]):
                original_length = (
                    (input_ids != self.tokenizer.pad_token_id).sum().item()
                )

                # Calculate QLEN dynamically for each query
                if original_length % 32 <= 8:
                    QLEN = original_length + 8
                else:
                    QLEN = ceil(original_length / 32) * 32

                if original_length < QLEN:
                    pad_length = QLEN - original_length
                    padded_input_ids = input_ids.tolist() + [mask_token_id] * pad_length
                    padded_attention_mask = (
                        encoding["attention_mask"][i].tolist() + [0] * pad_length
                    )
                else:
                    padded_input_ids = input_ids[:QLEN].tolist()
                    padded_attention_mask = encoding["attention_mask"][i][
                        :QLEN
                    ].tolist()

                new_encodings["input_ids"].append(padded_input_ids)
                new_encodings["attention_mask"].append(padded_attention_mask)

            for key in new_encodings:
                new_encodings[key] = torch.tensor(
                    new_encodings[key], device=self.device
                )

            encoding = new_encodings

        encoding = {key: value.to(self.device) for key, value in encoding.items()}
        return encoding

    def _query_encode(self, query: list[str]):
        return self._encode(query, self.query_token_id, is_query=True)

    def _document_encode(self, documents: list[str]):
        return self._encode(documents, self.document_token_id)

    def _to_embs(self, encoding) -> torch.Tensor:
        with torch.inference_mode():
            # embs = self.model(**encoding).last_hidden_state.squeeze(1)
            embs = self.model(**encoding)
        if self.normalize:
            embs = embs / embs.norm(dim=-1, keepdim=True)
        return embs

    def _rerank(self, query: str, documents: list[str]) -> list[float]:
        query_encoding = self._query_encode([query])
        documents_encoding = self._document_encode(documents)
        query_embeddings = self._to_embs(query_encoding)
        document_embeddings = self._to_embs(documents_encoding)
        scores = (
            _colbert_score(
                query_embeddings,
                document_embeddings,
                query_encoding["attention_mask"],
                documents_encoding["attention_mask"],
            )
            .cpu()
            .tolist()[0]
        )
        return scores


class ColBERTRanker(BaseRanker):
    def __init__(
        self,
        model_name: str,
        batch_size: int = 32,
        dtype: Optional[Union[str, torch.dtype]] = None,
        device: Optional[Union[str, torch.device]] = None,
        verbose: int = 1,
        query_token: str = "[unused0]",
        document_token: str = "[unused1]",
        **kwargs,
    ):
        self.verbose = verbose
        self.device = get_device(device, self.verbose)
        self.dtype = get_dtype(dtype, self.device, self.verbose)
        self.batch_size = batch_size
        vprint(
            f"Loading model {model_name}, this might take a while...",
            self.verbose,
        )
        tokenizer_kwargs = kwargs.get("tokenizer_kwargs", {})
        self.tokenizer = AutoTokenizer.from_pretrained(model_name, **tokenizer_kwargs)
        model_kwargs = kwargs.get("model_kwargs", {})
        self.model = (
            ColBERTModel.from_pretrained(
                model_name,
                **model_kwargs
            )
            .to(self.device)
            .to(self.dtype)
        )
        self.model.eval()
        self.query_max_length = 32  # Lower bound
        self.doc_max_length = (
            self.model.config.max_position_embeddings - 2
        )  # Upper bound
        self.query_token_id: int = self.tokenizer.convert_tokens_to_ids(query_token)  # type: ignore
        self.document_token_id: int = self.tokenizer.convert_tokens_to_ids(
            document_token
        )  # type: ignore
        self.normalize = True

    def rank(
        self,
        query: str,
        docs: Union[Document, str, List[Document], List[str]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
    ) -> RankedResults:
        docs = prep_docs(docs, doc_ids, metadata)

        scores = self._colbert_rank(query, [d.text for d in docs])
        ranked_results = [
            Result(document=doc, score=score, rank=idx + 1)
            for idx, (doc, score) in enumerate(
                sorted(zip(docs, scores), key=lambda x: x[1], reverse=True)
            )
        ]
        return RankedResults(results=ranked_results, query=query, has_scores=True)

    def score(self, query: str, doc: str) -> float:
        scores = self._colbert_rank(query, [doc])
        return scores[0] if scores else 0.0

    @torch.inference_mode()
    def _colbert_rank(
        self,
        query: str,
        docs: List[str],
    ) -> List[float]:
        query_encoding = self._query_encode([query])
        documents_encoding = self._document_encode(docs)
        query_embeddings = self._to_embs(query_encoding)
        document_embeddings = self._to_embs(documents_encoding)
        scores = (
            _colbert_score(
                query_embeddings,
                document_embeddings,
                query_encoding["attention_mask"],
                documents_encoding["attention_mask"],
            )
            .cpu()
            .tolist()[0]
        )
        return scores

    def _query_encode(self, query: list[str]):
        return self._encode(
            query, self.query_token_id, max_length=self.doc_max_length, is_query=True
        )

    def _document_encode(self, documents: list[str]):
        tokenized_doc_lengths = [
            len(
                self.tokenizer.encode(
                    doc, max_length=self.doc_max_length, truncation=True
                )
            )
            for doc in documents
        ]
        max_length = max(tokenized_doc_lengths)
        max_length = (
            ceil(max_length / 32) * 32
        )  # Round up to the nearest multiple of 32
        max_length = max(
            max_length, self.query_max_length
        )  # Ensure not smaller than query_max_length
        max_length = int(
            min(max_length, self.doc_max_length)
        )  # Ensure not larger than doc_max_length
        return self._encode(documents, self.document_token_id, max_length)

    def _encode(
        self,
        texts: list[str],
        insert_token_id: int,
        max_length: int,
        is_query: bool = False,
    ):
        encoding = self.tokenizer(
            texts,
            return_tensors="pt",
            padding=True,
            max_length=max_length - 1,  # for insert token
            truncation=True,
        )
        encoding = _insert_token(encoding, insert_token_id)  # type: ignore

        if is_query:
            mask_token_id = self.tokenizer.mask_token_id

            new_encodings = {"input_ids": [], "attention_mask": []}

            for i, input_ids in enumerate(encoding["input_ids"]):
                original_length = (
                    (input_ids != self.tokenizer.pad_token_id).sum().item()
                )

                # Calculate QLEN dynamically for each query
                if original_length % 16 <= 8:
                    QLEN = original_length + 8
                else:
                    QLEN = ceil(original_length / 16) * 16

                if original_length < QLEN:
                    pad_length = QLEN - original_length
                    padded_input_ids = input_ids.tolist() + [mask_token_id] * pad_length
                    padded_attention_mask = (
                        encoding["attention_mask"][i].tolist() + [0] * pad_length
                    )
                else:
                    padded_input_ids = input_ids[:QLEN].tolist()
                    padded_attention_mask = encoding["attention_mask"][i][
                        :QLEN
                    ].tolist()

                new_encodings["input_ids"].append(padded_input_ids)
                new_encodings["attention_mask"].append(padded_attention_mask)

            for key in new_encodings:
                new_encodings[key] = torch.tensor(
                    new_encodings[key], device=self.device
                )

            encoding = new_encodings

        encoding = {key: value.to(self.device) for key, value in encoding.items()}
        return encoding

    def _to_embs(self, encoding) -> torch.Tensor:
        with torch.inference_mode():
            batched_embs = []
            for i in range(0, encoding["input_ids"].size(0), self.batch_size):
                batch_encoding = {
                    key: val[i : i + self.batch_size] for key, val in encoding.items()
                }
                batch_embs = self.model(**batch_encoding)
                batched_embs.append(batch_embs)
            embs = torch.cat(batched_embs, dim=0)
        if self.normalize:
            embs = embs / embs.norm(dim=-1, keepdim=True)
        return embs



from typing import Optional, Union, List
from rerankers.models.ranker import BaseRanker
from rerankers.documents import Document
from rerankers.results import RankedResults, Result
from rerankers.utils import prep_docs

from rank_llm.data import Candidate, Query, Request
from rank_llm.rerank.vicuna_reranker import VicunaReranker
from rank_llm.rerank.zephyr_reranker import ZephyrReranker
from rank_llm.rerank.rank_gpt import SafeOpenai
from rank_llm.rerank.reranker import Reranker as rankllm_Reranker


class RankLLMRanker(BaseRanker):
    def __init__(
        self,
        model: str,
        api_key: Optional[str] = None,
        lang: str = "en",
        verbose: int = 1,
    ) -> "RankLLMRanker":
        self.api_key = api_key
        self.model = model
        self.verbose = verbose
        self.lang = lang

        if "zephyr" in self.model.lower():
            self.rankllm_ranker = ZephyrReranker()
        elif "vicuna" in self.model.lower():
            self.rankllm_ranker = VicunaReranker()
        elif "gpt" in self.model.lower():
            self.rankllm_ranker = rankllm_Reranker(
                SafeOpenai(model=self.model, context_size=4096, keys=self.api_key)
            )

    def rank(
        self,
        query: str,
        docs: Union[str, List[str], Document, List[Document]],
        doc_ids: Optional[Union[List[str], List[int]]] = None,
        metadata: Optional[List[dict]] = None,
        rank_start: int = 0,
        rank_end: int = 0,
    ) -> RankedResults:
        docs = prep_docs(docs, doc_ids, metadata)

        request = Request(
            query=Query(text=query, qid=1),
            candidates=[
                Candidate(doc={"text": doc.text}, docid=doc_idx, score=1)
                for doc_idx, doc in enumerate(docs)
            ],
        )

        rankllm_results = self.rankllm_ranker.rerank(
            request,
            rank_end=len(docs) if rank_end == 0 else rank_end,
            window_size=min(20, len(docs)),
            step=10,
        )

        ranked_docs = []

        for rank, result in enumerate(rankllm_results.candidates, start=rank_start):
            ranked_docs.append(
                Result(
                    document=docs[result.docid],
                    rank=rank,
                )
            )

        return RankedResults(results=ranked_docs, query=query, has_scores=False)

    def score(self):
        print("Listwise ranking models like RankLLM cannot output scores!")
        return None
